class
	BOOLEAN_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_if_statement,
			end_visit_while_statement,
			end_visit_unary_expression,
			end_visit_generic_expression,
			end_visit_jump_statement,
			end_visit_assignment_expression,
			end_visit_function_call,
			post_visit
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description: STRING
		do
			Result := "Transforming non-boolean (e.g. integer) conditionals to boolean ones"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_if_statement (node: AST_IF_STATEMENT)
		do
			transform_to_boolean (node.condition)
		end

	end_visit_while_statement (node: AST_WHILE_STATEMENT)
		do
			transform_to_boolean (node.condition)
		end

	end_visit_unary_expression (node: AST_UNARY_EXPRESSION)
		do
			if node.type = C_op_not then
				transform_to_boolean (node.expression)
			elseif
				node.type = C_op_minus and
				attached {AST_BASIC_TYPE} node.expression.type_binding as bt and then bt.type = Ast_bt_bool
			then
				transform_to_int (node.expression)
			end
		end

	end_visit_generic_expression (node: AST_GENERIC_EXPRESSION)
		local
			is_left_boolean, is_right_boolean: BOOLEAN
			bool_expr, non_bool_expr: AST_EXPRESSION
		do
			if node.is_logical_expression then
				-- `5 || 4' -> `(5).to_boolean || (4).to_boolean'
				transform_to_boolean (node.left)
				transform_to_boolean (node.right)
			else
				is_left_boolean := attached {AST_BASIC_TYPE} node.left.type_binding as tmp and then tmp.type = Ast_bt_bool
				is_right_boolean := attached {AST_BASIC_TYPE} node.right.type_binding as tmp and then tmp.type = Ast_bt_bool
				if
					node.is_equality_expression and
					((is_left_boolean and not is_right_boolean) or (is_right_boolean and not is_left_boolean))
				then
					if is_left_boolean then
						bool_expr := node.left
						non_bool_expr := node.right
					else
						bool_expr := node.right
						non_bool_expr := node.left
					end
					if
						attached {AST_DECIMAL_CONSTANT} non_bool_expr as tmp and then
						(tmp.value.is_equal ("0") or tmp.value.is_equal ("1"))
					then
						-- `a_bool_expr == 0' -> `!a_bool_expr'
						-- `a_bool_expr /= 0' -> `a_bool_expr'
						-- `a_bool_expr == 1' -> `a_bool_expr'
						-- `a_bool_expr /= 1' -> `!a_bool_expr'
						if
							(node.type = C_op_equal and tmp.value.is_equal ("0")) or
							(node.type = C_op_not_equal and tmp.value.is_equal ("1"))
						then
							replace_node (node, create_negated_expression (bool_expr))
						else
							replace_node (node, bool_expr.ast_twin)
						end
					else
						-- `(a == b) == an_expr' -> `(a == b) == (an_expr).to_boolean'
						transform_to_boolean (non_bool_expr)
					end
				end
			end
		end

	end_visit_jump_statement (node: AST_JUMP_STATEMENT)
		do
			if
				node.type = C_kw_return and then
				node.return_expression /= Void and then
				attached {AST_BASIC_TYPE} node.return_expression.type_binding as tmp and then tmp.type = Ast_bt_bool and then
				attached {AST_BASIC_TYPE} function_return_type (node) as ret_tmp and then ret_tmp.type = Ast_bt_int
			then
				transform_to_int (node.return_expression)
			end
		end

	end_visit_assignment_expression (node: AST_ASSIGNMENT_EXPRESSION)
		do
			if
				attached {AST_BASIC_TYPE} node.target.type_binding as tt and then tt.type = Ast_bt_int and then
				attached {AST_BASIC_TYPE} node.source.type_binding as st and then st.type = Ast_bt_bool
			then
				transform_to_int (node.source)
			end
		end

	end_visit_function_call (node: AST_FUNCTION_CALL)
		local
			src: LIST [AST_EXPRESSION]
			dest: LIST [AST_NODE]
			dest_type: AST_TYPE
		do
			if node.declaration_binding /= Void then
				src := node.parameters
				dest := node.declaration_binding.parameter_list.parameters
				from
					src.start
					dest.start
				until
					src.after or dest.after
				loop
					if attached {AST_VARIABLE_DECLARATION} dest.item as tmp then
						dest_type := tmp.type
					elseif attached {AST_TYPE} dest.item as tmp then
						dest_type := tmp
					end

					if
						(attached {AST_BASIC_TYPE} dest_type as dt and then dt.type = Ast_bt_int) and
						(attached {AST_BASIC_TYPE} src.item.type_binding as st and then st.type = Ast_bt_bool)
					then
						transform_to_int (src.item)
					end

					src.forth
					dest.forth
				end
			end
		end

	post_visit (node: AST_NODE)
		do
			if
				attached {AST_EXPRESSION} node as e and then
				attached {AST_BASIC_TYPE} e.type_binding as bt and then bt.type = Ast_bt_bool and then
				attached {AST_GENERIC_EXPRESSION} e.parent as ge and then
				(
					ge.is_additive_expression or ge.is_relational_expression or ge.is_bit_expression or
					ge.is_multiplicative_expression or ge.is_shift_expression
				)
			then
				-- `5 || 4' in `(5 || 4) + 3' -> `(5 || 4).to_integer'
				transform_to_int (e)
			end
		end

feature {NONE}

	transform_to_boolean (node: AST_EXPRESSION)
		require
			node /= Void
		do
			if attached {AST_DECIMAL_CONSTANT} node as decimal_constant then
				-- `if(1)'  ->  `if(True)'
				if decimal_constant.value.is_equal ("0") then
					replace_node (node, create {AST_BOOLEAN_CONSTANT}.make (False))
				else
					replace_node (node, create {AST_BOOLEAN_CONSTANT}.make (True))
				end
			elseif attached {AST_BASIC_TYPE} node.type_binding as tmp then
				if tmp.type = Ast_bt_bool then
					-- nothing to do
				elseif
					tmp.type = Ast_bt_char or tmp.type = Ast_bt_short or tmp.type = Ast_bt_int or tmp.type = Ast_bt_long or
					tmp.type = Ast_bt_long_long or tmp.type = Ast_bt_unsigned_char or tmp.type = Ast_bt_unsigned_short or
					tmp.type = Ast_bt_unsigned_int or tmp.type = Ast_bt_unsigned_long or tmp.type = Ast_bt_unsigned_long_long
				then
					-- `if(an_int_expression)'  ->  `if((an_int_expression).to_boolean)'
					replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (
							<< create {AST_PARENTHESIZED_EXPRESSION}.make (node.ast_twin), ".to_boolean">>,
							create {AST_BASIC_TYPE}.make (Ast_bt_bool)
						))
				elseif tmp.type = Ast_bt_float or tmp.type = Ast_bt_double then
					-- `if(a_real_expression)'  ->  `if((a_real_expression).truncated_to_integer.to_boolean)'
					replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (
							<< create {AST_PARENTHESIZED_EXPRESSION}.make (node.ast_twin), ".truncated_to_integer.to_boolean">>,
							create {AST_BASIC_TYPE}.make (Ast_bt_bool)
						))
				else
					log ("Unsupported condition type: " + node.type_binding.c_code)
				end
			elseif
				attached {AST_EXPRESSION} node.non_cast_expression as e and then
				(attached {AST_POINTER_TYPE} e.type_binding or attached {AST_ARRAY_TYPE} e.type_binding)
			then
				-- `if(a_pointer)'  ->  `if(not a_pointer.is_void)'
				replace_node (node, create {AST_PARENTHESIZED_EXPRESSION}.make (create {AST_UNARY_EXPRESSION}.make (
						C_op_not,
						create {AST_EIFFEL_EXPRESSION}.make_mixed (
								<<create {AST_PARENTHESIZED_EXPRESSION}.make (e.ast_twin), ".is_void">>,
								create {AST_BASIC_TYPE}.make (Ast_bt_bool)
							)
					)))
			elseif attached {AST_EIFFEL_STRING_TYPE} node.type_binding then
				-- `if(a_string)'  ->  `if(a_string /= Void)'
				replace_node (node, create {AST_PARENTHESIZED_EXPRESSION}.make (create {AST_GENERIC_EXPRESSION}.make (
						C_op_not_equal,
						node.ast_twin,
						create {AST_EIFFEL_EXPRESSION}.make ("Void", create {AST_BASIC_TYPE}.make (Ast_bt_void))
					)))
			else
				log ("Unsupported condition type: " + node.type_binding.c_code)
			end
		end

	transform_to_int (node: AST_EXPRESSION)
		require
			node /= Void and then attached {AST_BASIC_TYPE} node.type_binding as tmp and then tmp.type = Ast_bt_bool
		do
			-- `a_bool_expression'  ->  `(a_bool_expression).to_integer'
			replace_node (node, create {AST_EIFFEL_EXPRESSION}.make_mixed (
					<<create {AST_PARENTHESIZED_EXPRESSION}.make (node.ast_twin), ".to_integer">>,
					create {AST_BASIC_TYPE}.make (Ast_bt_int)
				))
		end

	function_return_type (func_child: AST_NODE): AST_TYPE
		require
			func_child /= Void
		local
			cur: AST_NODE
		do
			from
				cur := func_child.parent
			until
				cur = Void
			loop
				if attached {AST_FUNCTION_DECLARATION} cur as tmp then
					Result := tmp.return_type
					cur := Void
				else
					cur := cur.parent
				end
			end
		end

end
